@@ -1,23 +1,20 @@
 # XXX TypeErrors on calling handlers, or on bad return values from a
 # handler, are obscure and unhelpful.
 
-from io import BytesIO
-import abc
-import functools
 import os
 import platform
-import re
 import sys
 import sysconfig
-import textwrap
 import unittest
 import traceback
+from io import BytesIO
+from test import support
+from test.support import os_helper
 
 from xml.parsers import expat
 from xml.parsers.expat import errors
 
-from test.support import import_helper
-from test.support import sortdict
+from test.support import sortdict, is_emscripten, is_wasi
 
 
 class SetAttributeTest(unittest.TestCase):
@@ -446,33 +443,59 @@ class BufferTextTest(unittest.TestCase):
 # Test handling of exception from callback:
 class HandlerExceptionTest(unittest.TestCase):
     def StartElementHandler(self, name, attrs):
-        raise RuntimeError(name)
+        raise RuntimeError(f'StartElementHandler: <{name}>')
 
     def check_traceback_entry(self, entry, filename, funcname):
-        self.assertEqual(os.path.basename(entry[0]), filename)
-        self.assertEqual(entry[2], funcname)
+        self.assertEqual(os.path.basename(entry.filename), filename)
+        self.assertEqual(entry.name, funcname)
 
+    @support.cpython_only
     def test_exception(self):
+        # gh-66652: test _PyTraceback_Add() used by pyexpat.c to inject frames
+
+        # Change the current directory to the Python source code directory
+        # if it is available.
+        src_dir = sysconfig.get_config_var('abs_builddir')
+        if src_dir:
+            have_source = os.path.isdir(src_dir)
+        else:
+            have_source = False
+        if have_source:
+            with os_helper.change_cwd(src_dir):
+                self._test_exception(have_source)
+        else:
+            self._test_exception(have_source)
+
+    def _test_exception(self, have_source):
+        # Use path relative to the current directory which should be the Python
+        # source code directory (if it is available).
+        PYEXPAT_C = os.path.join('Modules', 'pyexpat.c')
+
         parser = expat.ParserCreate()
         parser.StartElementHandler = self.StartElementHandler
         try:
             parser.Parse(b"<a><b><c/></b></a>", True)
-            self.fail()
-        except RuntimeError as e:
-            self.assertEqual(e.args[0], 'a',
-                             "Expected RuntimeError for element 'a', but" + \
-                             " found %r" % e.args[0])
-            # Check that the traceback contains the relevant line in pyexpat.c
-            entries = traceback.extract_tb(e.__traceback__)
-            self.assertEqual(len(entries), 3)
-            self.check_traceback_entry(entries[0],
-                                       "test_pyexpat.py", "test_exception")
-            self.check_traceback_entry(entries[1],
-                                       "pyexpat.c", "StartElement")
-            self.check_traceback_entry(entries[2],
-                                       "test_pyexpat.py", "StartElementHandler")
-            if sysconfig.is_python_build() and not (sys.platform == 'win32' and platform.machine() == 'ARM'):
-                self.assertIn('call_with_frame("StartElement"', entries[1][3])
+
+            self.fail("the parser did not raise RuntimeError")
+        except RuntimeError as exc:
+            self.assertEqual(exc.args[0], 'StartElementHandler: <a>', exc)
+            entries = traceback.extract_tb(exc.__traceback__)
+
+        self.assertEqual(len(entries), 3, entries)
+        self.check_traceback_entry(entries[0],
+                                   "test_pyexpat.py", "_test_exception")
+        self.check_traceback_entry(entries[1],
+                                   os.path.basename(PYEXPAT_C),
+                                   "StartElement")
+        self.check_traceback_entry(entries[2],
+                                   "test_pyexpat.py", "StartElementHandler")
+
+        # Check that the traceback contains the relevant line in
+        # Modules/pyexpat.c. Skip the test if Modules/pyexpat.c is not
+        # available.
+        if have_source and os.path.exists(PYEXPAT_C):
+            self.assertIn('call_with_frame("StartElement"',
+                          entries[1].line)
 
 
 # Test Current* members:
@@ -825,199 +848,5 @@ class ReparseDeferralTest(unittest.TestCase):
         self.assertEqual(started, ['doc'])
 
 
-class AttackProtectionTestBase(abc.ABC):
-    """
-    Base class for testing protections against XML payloads with
-    disproportionate amplification.
-
-    The protections being tested should detect and prevent attacks
-    that leverage disproportionate amplification from small inputs.
-    """
-
-    @staticmethod
-    def exponential_expansion_payload(*, nrows, ncols, text='.'):
-        """Create a billion laughs attack payload.
-
-        Be careful: the number of total items is pow(n, k), thereby
-        requiring at least pow(ncols, nrows) * sizeof(text) memory!
-        """
-        template = textwrap.dedent(f"""\
-            <?xml version="1.0"?>
-            <!DOCTYPE doc [
-                <!ENTITY row0 "{text}">
-                <!ELEMENT doc (#PCDATA)>
-            {{body}}
-            ]>
-            <doc>&row{nrows};</doc>
-        """).rstrip()
-
-        body = '\n'.join(
-            f'<!ENTITY row{i + 1} "{f"&row{i};" * ncols}">'
-            for i in range(nrows)
-        )
-        body = textwrap.indent(body, ' ' * 4)
-        return template.format(body=body)
-
-    def test_payload_generation(self):
-        # self-test for exponential_expansion_payload()
-        payload = self.exponential_expansion_payload(nrows=2, ncols=3)
-        self.assertEqual(payload, textwrap.dedent("""\
-            <?xml version="1.0"?>
-            <!DOCTYPE doc [
-                <!ENTITY row0 ".">
-                <!ELEMENT doc (#PCDATA)>
-                <!ENTITY row1 "&row0;&row0;&row0;">
-                <!ENTITY row2 "&row1;&row1;&row1;">
-            ]>
-            <doc>&row2;</doc>
-        """).rstrip())
-
-    def assert_root_parser_failure(self, func, /, *args, **kwargs):
-        """Check that func(*args, **kwargs) is invalid for a sub-parser."""
-        msg = "parser must be a root parser"
-        self.assertRaisesRegex(expat.ExpatError, msg, func, *args, **kwargs)
-
-    @abc.abstractmethod
-    def assert_rejected(self, func, /, *args, **kwargs):
-        """Assert that func(*args, **kwargs) triggers the attack protection.
-
-        Note: this method must ensure that the attack protection being tested
-        is the one that is actually triggered at runtime, e.g., by matching
-        the exact error message.
-        """
-
-    @abc.abstractmethod
-    def set_activation_threshold(self, parser, threshold):
-        """Set the activation threshold for the tested protection."""
-
-    @abc.abstractmethod
-    def set_maximum_amplification(self, parser, max_factor):
-        """Set the maximum amplification factor for the tested protection."""
-
-    @abc.abstractmethod
-    def test_set_activation_threshold__threshold_reached(self):
-        """Test when the activation threshold is exceeded."""
-
-    @abc.abstractmethod
-    def test_set_activation_threshold__threshold_not_reached(self):
-        """Test when the activation threshold is not exceeded."""
-
-    def test_set_activation_threshold__invalid_threshold_type(self):
-        parser = expat.ParserCreate()
-        setter = functools.partial(self.set_activation_threshold, parser)
-
-        self.assertRaises(TypeError, setter, 1.0)
-        self.assertRaises(TypeError, setter, -1.5)
-        self.assertRaises(ValueError, setter, -5)
-
-    def test_set_activation_threshold__invalid_threshold_range(self):
-        _testcapi = import_helper.import_module("_testcapi")
-        parser = expat.ParserCreate()
-        setter = functools.partial(self.set_activation_threshold, parser)
-
-        self.assertRaises(OverflowError, setter, _testcapi.ULLONG_MAX + 1)
-
-    def test_set_activation_threshold__fail_for_subparser(self):
-        parser = expat.ParserCreate()
-        subparser = parser.ExternalEntityParserCreate(None)
-        setter = functools.partial(self.set_activation_threshold, subparser)
-        self.assert_root_parser_failure(setter, 12345)
-
-    @abc.abstractmethod
-    def test_set_maximum_amplification__amplification_exceeded(self):
-        """Test when the amplification factor is exceeded."""
-
-    @abc.abstractmethod
-    def test_set_maximum_amplification__amplification_not_exceeded(self):
-        """Test when the amplification factor is not exceeded."""
-
-    def test_set_maximum_amplification__infinity(self):
-        inf = float('inf')  # an 'inf' threshold is allowed by Expat
-        parser = expat.ParserCreate()
-        self.assertIsNone(self.set_maximum_amplification(parser, inf))
-
-    def test_set_maximum_amplification__invalid_max_factor_type(self):
-        parser = expat.ParserCreate()
-        setter = functools.partial(self.set_maximum_amplification, parser)
-
-        self.assertRaises(TypeError, setter, None)
-        self.assertRaises(TypeError, setter, 'abc')
-
-    def test_set_maximum_amplification__invalid_max_factor_range(self):
-        parser = expat.ParserCreate()
-        setter = functools.partial(self.set_maximum_amplification, parser)
-
-        msg = re.escape("'max_factor' must be at least 1.0")
-        self.assertRaisesRegex(expat.ExpatError, msg, setter, float('nan'))
-        self.assertRaisesRegex(expat.ExpatError, msg, setter, 0.99)
-
-    def test_set_maximum_amplification__fail_for_subparser(self):
-        parser = expat.ParserCreate()
-        subparser = parser.ExternalEntityParserCreate(None)
-        setter = functools.partial(self.set_maximum_amplification, subparser)
-        self.assert_root_parser_failure(setter, 123.45)
-
-
-@unittest.skipIf(expat.version_info < (2, 7, 2), "requires Expat >= 2.7.2")
-class MemoryProtectionTest(AttackProtectionTestBase, unittest.TestCase):
-
-    # NOTE: with the default Expat configuration, the billion laughs protection
-    # may hit before the allocation limiter if exponential_expansion_payload()
-    # is not carefully parametrized. As such, the payloads should be chosen so
-    # that either the allocation limiter is hit before other protections are
-    # triggered or no protection at all is triggered.
-
-    def assert_rejected(self, func, /, *args, **kwargs):
-        """Check that func(*args, **kwargs) hits the allocation limit."""
-        msg = r"out of memory: line \d+, column \d+"
-        self.assertRaisesRegex(expat.ExpatError, msg, func, *args, **kwargs)
-
-    def set_activation_threshold(self, parser, threshold):
-        return parser.SetAllocTrackerActivationThreshold(threshold)
-
-    def set_maximum_amplification(self, parser, max_factor):
-        return parser.SetAllocTrackerMaximumAmplification(max_factor)
-
-    def test_set_activation_threshold__threshold_reached(self):
-        parser = expat.ParserCreate()
-        # Choose a threshold expected to be always reached.
-        self.set_activation_threshold(parser, 3)
-        # Check that the threshold is reached by choosing a small factor
-        # and a payload whose peak amplification factor exceeds it.
-        self.assertIsNone(self.set_maximum_amplification(parser, 1.0))
-        payload = self.exponential_expansion_payload(ncols=10, nrows=4)
-        self.assert_rejected(parser.Parse, payload, True)
-
-    def test_set_activation_threshold__threshold_not_reached(self):
-        parser = expat.ParserCreate()
-        # Choose a threshold expected to be never reached.
-        self.set_activation_threshold(parser, pow(10, 5))
-        # Check that the threshold is reached by choosing a small factor
-        # and a payload whose peak amplification factor exceeds it.
-        self.assertIsNone(self.set_maximum_amplification(parser, 1.0))
-        payload = self.exponential_expansion_payload(ncols=10, nrows=4)
-        self.assertIsNotNone(parser.Parse(payload, True))
-
-    def test_set_maximum_amplification__amplification_exceeded(self):
-        parser = expat.ParserCreate()
-        # Unconditionally enable maximum activation factor.
-        self.set_activation_threshold(parser, 0)
-        # Choose a max amplification factor expected to always be exceeded.
-        self.assertIsNone(self.set_maximum_amplification(parser, 1.0))
-        # Craft a payload for which the peak amplification factor is > 1.0.
-        payload = self.exponential_expansion_payload(ncols=1, nrows=2)
-        self.assert_rejected(parser.Parse, payload, True)
-
-    def test_set_maximum_amplification__amplification_not_exceeded(self):
-        parser = expat.ParserCreate()
-        # Unconditionally enable maximum activation factor.
-        self.set_activation_threshold(parser, 0)
-        # Choose a max amplification factor expected to never be exceeded.
-        self.assertIsNone(self.set_maximum_amplification(parser, 1e4))
-        # Craft a payload for which the peak amplification factor is < 1e4.
-        payload = self.exponential_expansion_payload(ncols=1, nrows=2)
-        self.assertIsNotNone(parser.Parse(payload, True))
-
-
 if __name__ == "__main__":
     unittest.main()
